﻿import migurski.gis.MapRegistrationMarker;

/*
 * -----------------------------------------------------------------
 * Copyright (c) 2005 Michal Migurski. All rights reserved.
 * -----------------------------------------------------------------
 *
 * AffineTransformer.as
 * 
 * @desc Generates equations for transforming point in 2D space.
 *
 * @author Michal Migurski
 * @author Darren David
 *
 * CVS Information:
 * $Id: AffineTransformer.as,v 1.6 2005/07/06 16:20:37 darren Exp $
 * 
 * -----------------------------------------------------------------
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * -----------------------------------------------------------------
 */

class migurski.gis.AffineTransformer
{//}

	//{ Properties
	private var _markers : Array;

	// references for the generated functions
	private var transformX 	: Function;
	private var transformY 	: Function;
	private var untransformX : Function;
	private var untransformY : Function;
	//}

	//{ Constructor

	public function AffineTransformer()
	{
		this._markers = new Array();
	}

	//}


	//{ Public Methods

	public function toString() : String
	{
		return "AffineTransformer[]";
	}

	/**
	* Generate the transform equations
	*/
	public function calibrate() : Void
	{
		if ( this._markers.length == 3 )
		{
			this.deriveTransforms( this._markers );
		}
		else
		{
			throw new Error( this + ": You need exactly 3 MapRegistrationMarkers to perform calibration" )
		}
	}	
	
	public function registerMarker( marker : MapRegistrationMarker ) : Void
	{
		/*************************************************************
		*
		* Each marker must contain x/y coordinates; latitude and longtiude
		* are only used to help remember what location the point represents.
		* 
		* These coordinates represent the location of that point in projected 
		* space. By entering the x/y values for the marker, you are defining 
		* what the x/y in projected space is vs. what the _x/_y is in Flash space. 
		* The affine transform uses this data to generate an equation that
		* that will tranform any point in projected space to Flash space.
		*
		***************************************************************/
		
		// if x + y are defined, use them in favor of lat/long
		if( marker.x == undefined || marker.y == undefined) 
		{ 
			
			throw new Error("Marker must contain x and y: " + marker );
		}
		else
		{
			if ( _global.SHOW_DEBUG_OUTPUT ) 
				trace( this + ': registered: ' + marker );
	
			marker.setLabel( marker.x + ', ' + marker.y );
			
			this._markers.push(marker);
		}
	}
	
	/**
	* Transform a point from projected space to Flash space
	*/
	public function transform( x : Number, y : Number ) : Object
	{
		var o : Object = new Object();
		o.x = this.transformX( x, y );
		o.y = this.transformY( x, y );
		
		return o;
	}
	
	/**
	* Transform a point from Flash space to projected space
	*/
	public function untransform( x : Number, y : Number ) : Object
	{
		var o : Object = new Object();
		o.x = this.untransformX( x, y );
		o.y = this.untransformY( x, y );
		
		return o;
	}	

	//}
	
	
	//{ Private Methods

	/**
	 * solves a system of linear equations:
	 *   z1 = (a * x1) + (b + y1) + c
	 *   z2 = (a * x2) + (b + y2) + c
	 *   z3 = (a * x3) + (b + y3) + c
	 *
	 * x1 - z3 are the known values.
	 * a, b, c are the unknowns to be solved.
	 * generates a function which will return z for any x, y.
	 * ask mike about pages 78-79 of his notebook for more info.
	 *
	 *  @param  array       knowns  array of known parameters.
	 *                              assumes that elements
	 *                              x1, x2, x3, y1, y2, y3, z1, z2, z3
	 *                              are all numeric.
	 *
	 *  @return function    returns a function that takes two arguments
	 *                      x & y and returns the corresponding z, based
	 *                      on a solution from the above knowns.
	 */
	private function linearSolution( knowns )
	{
		var unknowns = new Object();
		
		unknowns.a = (((knowns.z2 - knowns.z3) * (knowns.y1 - knowns.y2)) - ((knowns.z1 - knowns.z2) * (knowns.y2 - knowns.y3))) /
					 (((knowns.x2 - knowns.x3) * (knowns.y1 - knowns.y2)) - ((knowns.x1 - knowns.x2) * (knowns.y2 - knowns.y3)));
		
		unknowns.b = (((knowns.z2 - knowns.z3) * (knowns.x1 - knowns.x2)) - ((knowns.z1 - knowns.z2) * (knowns.x2 - knowns.x3))) /
					 (((knowns.y2 - knowns.y3) * (knowns.x1 - knowns.x2)) - ((knowns.y1 - knowns.y2) * (knowns.x2 - knowns.x3)));
				   
		unknowns.c = knowns.z1 - (knowns.x1 * unknowns.a) - (knowns.y1 * unknowns.b);
		
		if ( _global.SHOW_DEBUG_OUTPUT ) 
		{
			trace('\n'+this+":");
			trace('--> linear_solution:');
			trace('    x1, y1, z1: '+knowns.x1+', '+knowns.y1+', '+knowns.z1);
			trace('    x2, y2, z2: '+knowns.x2+', '+knowns.y2+', '+knowns.z2);
			trace('    x3, y3, z3: '+knowns.x3+', '+knowns.y3+', '+knowns.z3);
			trace('    a, b, c: '+unknowns.a+', '+unknowns.b+', '+unknowns.c);
		}
		
		var z1 = (unknowns.a * knowns.x1) + (unknowns.b * knowns.y1) + unknowns.c;
		if ( _global.SHOW_DEBUG_OUTPUT ) trace( this + ': z1 vs. knowns.z1: '+z1+', '+knowns.z1);
		if(Math.abs(z1 - knowns.z1) > 0.00001) { trace(this + ': failed on z1'); return false; }
		
		var z2 = (unknowns.a * knowns.x2) + (unknowns.b * knowns.y2) + unknowns.c;
		if ( _global.SHOW_DEBUG_OUTPUT ) trace(this + ': z2 vs. knowns.z2: '+z2+', '+knowns.z2);
		if(Math.abs(z2 - knowns.z2) > 0.00001) { trace(this + ': failed on z2'); return false; }
		
		var z3 = (unknowns.a * knowns.x3) + (unknowns.b * knowns.y3) + unknowns.c;
		if ( _global.SHOW_DEBUG_OUTPUT ) trace(this + ': z3 vs. knowns.z3: '+z3+', '+knowns.z3);
		if(Math.abs(z3 - knowns.z3) > 0.00001) { trace(this + ': failed on z3'); return false; }
		
		var f = function(x, y) 
		{
			if ( _global.SHOW_DEBUG_OUTPUT ) trace(this + ': anon function a, b, c: '+unknowns.a+', '+unknowns.b+', '+unknowns.c);
			return (unknowns.a * x) + (unknowns.b * y) + unknowns.c;
		};
	
		return f;    
	}
	
	/** 
	 * generates a set of 2D transformation functions.
	 * given a set of three well-chosen points (i.e., not on a
	 * line or in one place), each with properties y & x
	 * and placed at point _x, _y, returns a set
	 * of four functions useful for performing linear
	 * transformation from input (x/y or lon/lat) coordinates
	 * to output (_x/_y) coordinates, and back.
	 *
	 *  @param  array   markers     array of three movie clips.
	 *                              each must have input properties x, y defined.
	 *
	 *  @return array   array of functions:
	 *                  in2outx(x, y) returns output x for given input coords
	 *                  in2outy(x, y) returns output y for given input coords
	 *                  out2iny(x, y) returns input y for given output coords
	 *                  out2inx(x, y) returns input x for given output coords
	 */
	private function deriveTransforms( markers : Array )
	{
		var knowns = new Object();
	
		knowns.transformX 	= new Object();
		knowns.transformY 	= new Object();
		knowns.untransformY = new Object();
		knowns.untransformX = new Object();
	
		var i = 1;
		
		for( var m in markers) 
		{
			var marker : MapRegistrationMarker = markers[m];
			
			if ( _global.SHOW_DEBUG_OUTPUT ) 
				trace( this + ': ' + marker+', in x/y: '+marker.x+'/'+marker.y+' <--> out _x/_y: '+marker._x+'/'+marker._y);
	
			knowns.transformX['x'+i] = marker.x;
			knowns.transformX['y'+i] = marker.y;
			knowns.transformX['z'+i] = marker._x;
	
			knowns.transformY['x'+i] = marker.x;
			knowns.transformY['y'+i] = marker.y;
			knowns.transformY['z'+i] = marker._y;
	
			knowns.untransformY['x'+i] = marker._x;
			knowns.untransformY['y'+i] = marker._y;
			knowns.untransformY['z'+i] = marker.y;
	
			knowns.untransformX['x'+i] = marker._x;
			knowns.untransformX['y'+i] = marker._y;
			knowns.untransformX['z'+i] = marker.x;
			
			i += 1;
		}
		
		this.transformX 	= linearSolution( knowns.transformX );
		this.transformY 	= linearSolution( knowns.transformY );
		this.untransformY 	= linearSolution( knowns.untransformY );
		this.untransformX 	= linearSolution( knowns.untransformX );
	}

	//}
}
