﻿import migurski.gis.projections.ILinearProjection;
import migurski.util.MathUtil;

/*
 * -----------------------------------------------------------------
 * Copyright (c) 2005 Michal Migurski. All rights reserved.
 * -----------------------------------------------------------------
 *
 * LambertAzimuthalEqualAreaProjection.as
 * 
 * @desc calculates Lambert Azimuthal Equal-Area Projection
 *       http://mathworld.wolfram.com/LambertAzimuthalEqual-AreaProjection.html
 *
 * @author Michal Migurski
 * @author Darren David
 *
 * CVS Information:
 * $Id: LambertAzimuthalEqualAreaProjection.as,v 1.5 2005/07/04 17:32:51 darren Exp $
 * 
 * -----------------------------------------------------------------
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * -----------------------------------------------------------------
 */


class migurski.gis.projections.LambertAzimuthalEqualAreaProjection
implements ILinearProjection
{//}
	// these numbers are stored as radians
	private var _lat0 : Number;
	private var _lon0 : Number;
	
	
	/**
	* Constructor
	* @param latOrigin The latitudinal origin of the map in degrees
	* @param longOrigin The longitudinal origin of the map in degrees
	*/
	public function LambertAzimuthalEqualAreaProjection( lat0 : Number, lon0 : Number )
	{
		this._lat0 = MathUtil.deg2rad( lat0 );
		this._lon0 = MathUtil.deg2rad( lon0 );
	}
	
	
	/**
	 * Given lat/long in degrees in GPS space, projects point in Map space
	 * @param lat The latitude component of the point in degrees
	 * @param long The longitude component of the point in degrees
	 * @return An Object containing x and y properties
	 */
	public function GPSToMapCoords( lat : Number, long : Number) : Object
	{
		var point : Object = new Object();
		var latRadians : Number = MathUtil.deg2rad( lat );
		var longRadians : Number = MathUtil.deg2rad( long );
		
		var k : Number = Math.sqrt(2 / (1 + (Math.sin( this._lat0 ) * Math.sin( latRadians ) ) + ( Math.cos( this._lat0 ) * Math.cos( latRadians ) * Math.cos( longRadians - this._lon0))));
        
		point.x = k * Math.cos( latRadians ) * Math.sin( longRadians - this._lon0 );
		point.y = k * ((Math.cos( this._lat0 ) * Math.sin( latRadians )) - (Math.sin( this._lat0 ) * Math.cos( latRadians ) * Math.cos( longRadians - this._lon0)));
		
		return point;
	}
	

	/**
	 * Given x/y in Map space, convert to lat/long in degrees in GPS space
	 * @param lat The x component of the point in pixels
	 * @param long The y component of the point in pixels
	 * @return An Object containing lat and long properties
	 */
	public function MapToGPSCoords( x : Number, y : Number ) : Object
	{
		var coords : Object = new Object();
		var p : Number = Math.sqrt(Math.pow(x, 2) + Math.pow(y, 2));
		var c : Number = 2 * Math.asin(0.5 * p);
		
		if ( p == 0 ) 
		{
			coords.lat  = MathUtil.rad2deg( this._lat0 );
			coords.long = MathUtil.rad2deg( this._lon0 );
		} 
		else 
		{
			coords.lat  = MathUtil.rad2deg( Math.asin((Math.cos(c) * Math.sin(this._lat0)) + ((y * Math.sin(c) * Math.cos(this._lat0)) / p)) );
			coords.long = MathUtil.rad2deg( this._lon0 + Math.atan((x * Math.sin(c)) / ((p * Math.cos(this._lat0) * Math.cos(c)) - (y * Math.sin(this._lat0) * Math.sin(c)))) );
		}

		return coords;
	}
}